"use strict";
// This file is part of cwait, copyright (c) 2015-2017 BusFaster Ltd.
// Released under the MIT license, see LICENSE.
Object.defineProperty(exports, "__esModule", { value: true });
var BinaryHeap_1 = require("cdata/dist/BinaryHeap");
var Task_1 = require("./Task");
var TaskQueue = (function () {
    function TaskQueue(Promise, concurrency) {
        var _this = this;
        this.busyCount = 0;
        this.backlog = new BinaryHeap_1.BinaryHeap(function (a, b) { return a.stamp - b.stamp; });
        this.Promise = Promise;
        this.concurrency = concurrency;
        this.nextBound = function () { return _this.next(); };
        this.nextTickBound = function () {
            _this.timerStamp = 0;
            _this.next();
        };
    }
    /** Add a new task to the queue.
      * It will start when the number of other concurrent tasks is low enough.
      * @param func Function to call.
      * @param delay Initial delay in milliseconds before making the call. */
    TaskQueue.prototype.add = function (func, delay) {
        if (delay === void 0) { delay = 0; }
        if (this.busyCount < this.concurrency && !delay) {
            // Start the task immediately.
            ++this.busyCount;
            return (Task_1.tryFinally(func, this.nextBound, this.Promise));
        }
        else {
            // Schedule the task and return a promise resolving
            // to the result of task.start().
            var stamp = new Date().getTime() + delay;
            var task = new Task_1.Task(func, this.Promise, stamp);
            this.backlog.insert(task);
            return (task.delay());
        }
    };
    /** Consider current function idle until promise resolves.
      * Useful for making recursive calls. */
    TaskQueue.prototype.unblock = function (promise) {
        var _this = this;
        this.next();
        var onFinish = function () { return ++_this.busyCount; };
        promise.then(onFinish, onFinish);
        return (promise);
    };
    /** Wrap a function returning a promise, so that before running
      * it waits until concurrent invocations are below this queue's limit. */
    TaskQueue.prototype.wrap = function (func, thisObject) {
        var _this = this;
        return (function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return _this.add(function () { return func.apply(thisObject, args); });
        });
    };
    /** Start the next task from the backlog. */
    TaskQueue.prototype.next = function () {
        var stamp = new Date().getTime();
        var task = null;
        // If another task is eligible to run, get the next scheduled task.
        if (this.busyCount <= this.concurrency)
            task = this.backlog.peekTop();
        if (task && task.stamp <= stamp) {
            // A task remains, scheduled to start already. Resume it.
            task = this.backlog.extractTop();
            task.resume(this.nextBound);
        }
        else {
            // No tasks are scheduled to start right now.
            // Any new task can start immediately.
            --this.busyCount;
            if (task && (!this.timerStamp || task.stamp < this.timerStamp)) {
                // There is a task scheduled after a delay,
                // and no current timer firing before the delay.
                if (this.timerStamp) {
                    // There is a timer firing too late. Remove it.
                    clearTimeout(this.timer);
                }
                // Start a timer to clear timerStamp and call this function.
                this.timer = setTimeout(this.nextTickBound, stamp - task.stamp);
                this.timerStamp = task.stamp;
            }
        }
    };
    return TaskQueue;
}());
exports.TaskQueue = TaskQueue;
